#!/usr/bin/env python3
"""
This example illustrates how to perform constant-bandwidth univariate
kernel density estimation (KDE) with a brute-force calculation. The KDE
bandwidth is chosen by cross-validation (CV). This kind of calculation
works well only for relatively small sample sizes as the CV computational
complexity increases in proportion to the sample size squared.
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.2"
__date__ ="Sep 20 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
from npstat_density_estimation import histoPluginBandwidth1D
import matplotlib.pyplot as plt

# Some parameters for the script
sampleSize = 100   # Number of points to generate (sample size)
xmin = -3.5        # Left edge of the sample histogram
xmax = 3.5         # Right edge of the sample histogram
numBins = 35       # Number of bins for the sample histogram
scanPoints = 701   # Number of points to use for density scans
symbetaPower = 4   # Power parameter of the symmetric beta function used as
                   # the kernel. Use -1 to create a Gaussian kernel instead.
filterDegree = 0   # Degree of polynomial used to generate high-order kernels.
                   # In general, even numbers make the most sense here.
                   # Kernel order will then be filterDegree + 2. The kernel
                   # will be a bona-fide density (i.e., non-negative
                   # everywhere) only if filterDegree is 0.

# We will use a 1-d Gassian mixture to generate the random sample. The
# components of such a mixture are described by the "GaussianMixtureEntry"
# class. GaussianMixtureEntry constructor takes three arguments: the
# weight of the component, the mean of the Gaussian, and the standard
# deviation. Weights can not be negative and the standard deviations
# must be positive. The sum of the weights will be normalized internally.
component1 = ns.GaussianMixtureEntry(3.0, -1.0, 0.5)
component2 = ns.GaussianMixtureEntry(7.0,  1.0, 0.5)
mixture = ns.GaussianMixture1D(0.0, 1.0, (component1, component2))

# Generate a random sample according to the Gassian mixture density
rng = ns.MersenneTwister()
sample = mixture.generate(rng, sampleSize)

# Histogram this random sample
h1 = HistoND("int", ns.HistoAxis(numBins, xmin, xmax, "X"),
             "Sample Histogram", "Counts")
h1.fill(sample, 1)

# Create the kernel that will be used to perform KDE
kernel = ns.KDE1DHOSymbetaKernel(symbetaPower, filterDegree)

# Create the object which will be used to estimate the density
kde = ns.DoubleKDE1D(kernel, sample)

# Make an initial guess for the KDE bandwidth. It will use a robust
# measure of the sample scale, estimated from the histogram.
bw0 = histoPluginBandwidth1D(h1, sampleSize, filterDegree, symbetaPower)
print("Plug-in bandwidth is {:.4f}".format(bw0))

# Using the initial bandwidth guess, choose bandwidth by cross-validation.
# We can use some numerical optimization method (e.g., scipy.optimize), but
# here we will perform just a simple bandwidth scan. Note that for brute-force
# KDE the time of cross-validation calculations scales as O(sampleSize^2).
bwValues = ns.EquidistantInLogSpace(bw0/4.0, bw0*2.0, 50)

# Run the regularized likelihood cross-validation (RLCV).
# For the description of this method, see section 4.2 in
# https://doi.org/10.1080/10485252.2017.1371715
regularisationParameter = 0.5
rlcvFunctor = kde.rlcvFunctor(regularisationParameter)
rlcv, rlcvBandwidth = max([(rlcvFunctor(bw), bw) for bw in bwValues])
print("RLCV bandwidth is {:.4f}".format(rlcvBandwidth))

# Run the least squares cross-validation (LSCV). For this type of CV,
# the code will need to calculate the integral of density estimate squared.
# The parameters of the "lscvFunctor" method specify how this integration
# is performed. The method signature is
#
# lscvFunctor(xmin, xmax, nSubIntervals, nPoints)
#
# The interval of integration will be [xmin, xmax]. This interval will
# be split into "nSubIntervals" subintervals, and the Gauss-Legendre
# quadrature rule with "nPoints" points will be utilized on each subinterval.
lscvFunctor = kde.lscvFunctor(2*xmin, 2*xmax, 1, 256)
lscv, lscvBandwidth = max([(lscvFunctor(bw), bw) for bw in bwValues])
print("LSCV bandwidth is {:.4f}".format(lscvBandwidth))

# Plot the sample histogram
xh, yh = ns.histoOutline1D(h1)
plt.plot(xh, yh, 'k')

# Overlay the original density. Scale all densities so that their
# plots are compatible with the histogram height.
verticalScale = h1.binVolume()*sampleSize
xdens, ydens = ns.scanDensity1D(mixture, xmin, xmax, scanPoints)
plt.plot(xdens, ydens*verticalScale, 'g', label='True density')

# Overlay the KDE results with various bandwidth values
bandwidths = (bw0, rlcvBandwidth, lscvBandwidth)
colors     = ('c',       'r',          'b'     )
labels     = ('Plug-in', 'RLCV',       'LSCV'  )
for bw, color, l in zip(bandwidths, colors, labels):
    x, y = ns.scanKDE1D(kde, xmin, xmax, scanPoints, bw)
    plt.plot(x, y*verticalScale, color, label="{} estimate".format(l))

# Add labels and comments
plt.xlabel('X')
plt.ylabel('Density estimate')
plt.title('Brute-Force KDE')
plt.legend(loc=2)

# Display the window
plt.show()
