#!/usr/bin/env python3
"""
This example illustrates how to perform orthogonal series density
estimation (OSDE) in 1-d with a brute-force calculation. This kind
of calculation works well for relatively small sample sizes when
the density is expected to be smooth.
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Sep 23 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
from npstat_density_estimation import correctDensityEstimateGHU
import matplotlib.pyplot as plt
import numpy as np

# Some parameters for the script
sampleSize = 100   # Number of points to generate (sample size)
xmin = -3.5        # Left edge of the sample histogram
xmax = 3.5         # Right edge of the sample histogram
numBins = 35       # Number of bins for the sample histogram
nScanPoints = 700  # Number of points to use for density scans
maxDegree = 50     # Maximum polynomial degree to consider

# We will use a 1-d Gassian mixture to generate the random sample.
# See the "brute_force_kde_1d.py" script for more comments on this.
component1 = ns.GaussianMixtureEntry(3.0, -1.0, 0.5)
component2 = ns.GaussianMixtureEntry(7.0,  1.0, 0.5)
mixture = ns.GaussianMixture1D(0.0, 1.0, (component1, component2))

# Generate a random sample according to the Gassian mixture density
rng = ns.MersenneTwister()
sample = mixture.generate(rng, sampleSize)

# Histogram this random sample
h1 = HistoND("int", ns.HistoAxis(numBins, xmin, xmax, "X"),
             "Sample Histogram", "Counts")
h1.fill(sample, 1)

# Create the object which will be used to estimate the density.
# It is convenient to utilize Legendre polynomials for density
# estimation on closed intervals.
osde = ns.OSDE1D(ns.LegendreOrthoPoly1D(), xmin, xmax)

# Estimate polynomial expansion coefficients and their variances
coeffs = osde.sampleCoeffs(sample, maxDegree+1)
vars = osde.sampleCoeffsVariance(sample, coeffs, len(coeffs))

# Estimate series truncation point according to the Hart scheme
degHart = osde.optimalDegreeHart(coeffs, vars)
print("Hart scheme degree is", degHart)

# Truncate the series
hartCoeffs = np.resize(coeffs, degHart+1)

# Scan the OSDE density estimate
xosde, yosde = osde.seriesScan(hartCoeffs, nScanPoints)

# OSDE densities often become negative somewhere. Correct for this.
# The correcion method comes from the paper by Ingrid K. Glad,
# Nils Lid Hjort and Nikolai G. Ushakov, Scandinavian Journal of
# Statistics, Vol. 30, No. 2 (2003), pp. 415-427.
cellWidth = (osde.xmax() - osde.xmin())/len(yosde)
ycorr = correctDensityEstimateGHU(yosde, cellWidth)

# Plot the sample histogram
xh, yh = ns.histoOutline1D(h1)
plt.plot(xh, yh, 'k')

# Overlay the original density. Scale all densities so that their
# plots are compatible with the histogram height.
verticalScale = h1.binVolume()*sampleSize
xdens, ydens = ns.scanDensity1D(mixture, xmin, xmax, nScanPoints+1)
plt.plot(xdens, ydens*verticalScale, 'g', label='True density')

# Overlay the OSDE result
plt.plot(xosde, yosde*verticalScale, 'r', label='OSDE estimate')
plt.plot(xosde, ycorr*verticalScale, 'b', label='Corrected OSDE')

# Add labels and comments
plt.xlabel('X')
plt.ylabel('Density estimate')
plt.title('Brute-Force OSDE')
plt.legend(loc=2)

# Display the window
plt.show()
