#!/usr/bin/env python3
"""
This example illustrates how to perform local polynomial least squares
regression in 2-d with NPStat
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Oct 09 2019"

# Perform necessary imports
import math
import npstat as ns
from npstat_utils import makePointsFromFlat, BoxND, scanFunctorND
from npstat_local_regression import LocalQuadraticLeastSquaresND
import matplotlib.pyplot as plt
from npstat_plotting import colormeshNumpyArray
from mpl_toolkits.mplot3d import Axes3D

# Some parameters for the script
sampleSize = 200      # Number of points to generate (sample size).
regressionDegree = 1  # Degree of local polynomial used in regression.
                      # Can be 0, 1, or 2.
nDiscrete = 256       # Number of points for various discretizations.
sampleSigma = 0.3     # The standard deviation for the regression
                      # data (homoscedastic problem in this example).
xmin = 0.0            #
xmax = 2.0            # Support of the regression curve.
ymin = 0.0            #  
ymax = 1.0            #

# We will model the true mean of the regression curve
# with a product of sines
class TrueMean:
    def __init__(self, amplitude, kx, ky):
        self.a = amplitude
        self.kx = kx
        self.ky = ky
    def __call__(self, x, y):
        return self.a*math.sin(self.kx*x)*math.sin(self.ky*y)

mean = TrueMean(1.0, math.pi, 2.0*math.pi)

# Generate a sample of x, y coordinates flat on the support
rng = ns.MersenneTwister()
uniform = ns.UniformND((xmin, ymin), (xmax-xmin, ymax-ymin))
xysample = uniform.generate(rng, sampleSize)

# Generate the regression sample
zvalues = [ns.Gauss1D(mean(x, y), sampleSigma).random(rng) \
           for x, y in zip(xysample[0::2], xysample[1::2])]

# Build the local regression object. We will use a product kernel
# in this example. We also need to restructure the flat sample
# "xysample" into a set of 2-d points for consumption by the local
# regression code.
kernel = ns.ProductSymmetricBetaND((0.0, 0.0), (1.0, 1.0), 4)
xysample2d = makePointsFromFlat(2, xysample)
localReg = LocalQuadraticLeastSquaresND(regressionDegree, kernel,
                                        xysample2d, zvalues)

# Try a number of different bandwidth settings and select the best one
# by LSCV. We don't use AIC-based bandwidth selectors in this example
# as they become too slow for large samples.
referenceBandwidth = (2.0, 1.0)
bwFactors = ns.EquidistantInLogSpace(0.15, 0.6, 50)
bestBwLSCV = localReg.selectLSCVBandwidth(bwFactors, referenceBandwidth)
print("LSCV bandwidth is {:.5f}, {:.5f}".format(*bestBwLSCV))

# Curve for displaying the true mean
box = BoxND((xmin, xmax), (ymin, ymax))
zdense = scanFunctorND(mean, box, (nDiscrete, nDiscrete))

# Regression curve
zreg = scanFunctorND(localReg, box, (nDiscrete, nDiscrete))

# Plot what we have so far
fig = plt.figure()
a0 = fig.add_subplot(221)
mesh = colormeshNumpyArray(a0, zdense, box,
                           flipy=True, title="True Mean")
fig.colorbar(mesh, ax=a0)

a1 = fig.add_subplot(222, projection='3d')
a1.set_xlabel('X')
a1.set_ylabel('Y')
a1.set_zlabel('Z')
a1.set_title("Regression Points")
a1.scatter(xysample[0::2], xysample[1::2], zvalues, marker='.')

a2 = fig.add_subplot(223)
mesh = colormeshNumpyArray(a2, zreg, box,
                           flipy=True, title="Regression Curve")
fig.colorbar(mesh, ax=a2)

a3 = fig.add_subplot(224)
mesh = colormeshNumpyArray(a3, zreg - zdense, box,
                           flipy=True, title="Regression - True")
fig.colorbar(mesh, ax=a3)

# Display the window
if regressionDegree == 0:
    fig.suptitle('Local Average Regression (Nadaraya–Watson)')
elif regressionDegree == 1:
    fig.suptitle('Local Linear Regression')
elif regressionDegree == 2:
    fig.suptitle('Local Quadratic Regression')
fig.tight_layout(rect=[0, 0, 1, 0.95])
plt.show()
