#!/usr/bin/env python3
"""
This variation of plot_histo_1d_matplotlib.py script shows how to visualize
a histogram with Matplotlib using a bar plot
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Sep 13 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
import matplotlib.pyplot as plt
import numpy as np

# Create the histogram axis (needed to construct the histogram below).
# Here, we use an axis with uniform binning. The axis constructor arguments
# are the number of bins, left histogram edge, right histogram edge, and
# the axis label.
histoAxis = ns.HistoAxis(20, -5, 5, "X")

# The following line creates a 1-d histogram with bins of type "int"
h1 = HistoND("int", histoAxis, "Example NPStat Histogram", "Counts")

# We will fill the histogram with Gaussian random numbers. In order to do
# that, we need a random number generator and a statistical distribution
# object which will be used to convert random numbers on [0, 1) into the
# random numbers distributed as we like. The default constructor of
# MersenneTwister generator picks up a different random seed every time,
# so that the histogram will be different every time you run this script.
rng = ns.MersenneTwister()
d1 = ns.Gauss1D(0, 1)

# Fill the histogram with 1000 random numbers. The fastest way
# to do it, if you don't need to use the random sample for any
# other purpose, is via the NPStat "randomHistoFill1D" function.
ns.randomHistoFill1D(h1, rng, d1, 1000)

# Create Matplotlib plot window
fig = plt.figure()
ax = fig.add_subplot(111)

# The "plot_histo_1d.py" script obtained the histogram outline and employed
# the statement "ax.fill(x, y)" to draw it. Here, we will be drawing bars
# which can be obtained with the "histoBars" method of the NPStat package.
xscale = 0.6   # Horizontal scale for the bin width
xshift = 0.0   # Horizontal shift in the units of bin width
yscale = 1.0   # Vertical scale factor for the bars
yshift = []    # Vertical shifts for the bars. Empty list means no shifts.
left, height, width, bottom = ns.histoBars(h1, xscale, xshift, yscale, yshift)
ax.bar(left, height, width, bottom, color='b', lw=0)

# Set the x and y ranges of the plot
ax.set_xlim(h1.axis(0).min(), h1.axis(0).max())
ax.set_ylim(0, 1.1*np.max(np.array(bottom) + np.array(height)))

# Set the axis labels and the plot title
plt.xlabel(h1.axis(0).label())
plt.ylabel(h1.accumulatedDataLabel())
plt.title(h1.title())

# Display the window. The code enters Matplotlib event loop here.
# You will be able to use the window interface to manipulate the
# figure (pan, zoom) and to save it in various graphical formats.
plt.show()
