#include "geners/binaryIO.hh"
#include "geners/IOException.hh"

#include "npstat/stat/AllSymbetaParams1D.hh"

namespace npstat {
    AllSymbetaParams1D::AllSymbetaParams1D(
        int symbetaPower, double bandwidth, double degree,
        unsigned nbins, double binwidth,
        const BoundaryHandling& bm,
        unsigned excludedBin, bool excludeCentralPoint)
        : bandwidth_(bandwidth), degree_(degree), binwidth_(binwidth),
          nbins_(nbins), excludedBin_(excludedBin),
          symbetaPower_(symbetaPower),
          bm_(bm),
          excludeCentralPoint_(excludeCentralPoint)
    {
        if (bandwidth_ <= 0.0) throw std::invalid_argument(
            "In npstat::AllSymbetaParams1D constructor: "
            "bandwidth must be positive");
        if (degree_ < 0.0) throw std::invalid_argument(
            "In npstat::AllSymbetaParams1D constructor: "
            "degree parameter can not be negative");
        if (binwidth_ <= 0.0) throw std::invalid_argument(
            "In npstat::AllSymbetaParams1D constructor: "
            "bin width must be positive");
        if (!nbins_) std::invalid_argument(
            "In npstat::AllSymbetaParams1D constructor: "
            "number of bins must be positive");
    }

    bool AllSymbetaParams1D::write(std::ostream& of) const
    {
        gs::write_pod(of, bandwidth_);
        gs::write_pod(of, degree_);
        gs::write_pod(of, binwidth_);
        gs::write_pod(of, nbins_);
        gs::write_pod(of, excludedBin_);
        gs::write_pod(of, symbetaPower_);
        gs::write_obj(of, bm_);
        unsigned char exclude = excludeCentralPoint_;
        gs::write_pod(of, exclude);
        return !of.fail();
    }

    void AllSymbetaParams1D::restore(const gs::ClassId& id, std::istream& in,
                                     AllSymbetaParams1D* ptr)
    {
        static const gs::ClassId myClassId(
            gs::ClassId::makeId<AllSymbetaParams1D>());
        myClassId.ensureSameId(id);

        assert(ptr);

        gs::read_pod(in, &ptr->bandwidth_);
        gs::read_pod(in, &ptr->degree_);
        gs::read_pod(in, &ptr->binwidth_);
        gs::read_pod(in, &ptr->nbins_);
        gs::read_pod(in, &ptr->excludedBin_);
        gs::read_pod(in, &ptr->symbetaPower_);
        gs::restore_obj(in, &ptr->bm_);
        unsigned char exclude = 0;
        gs::read_pod(in, &exclude);
        ptr->excludeCentralPoint_ = exclude;

        if (in.fail()) throw gs::IOReadFailure(
            "In npstat::AllSymbetaParams1D::restore: "
            "input stream failure");
    }

    bool AllSymbetaParams1D::operator==(const AllSymbetaParams1D& r) const
    {
        if (bandwidth_ < 0.0 && r.bandwidth_ < 0.0)
            return true;

        unsigned ebin = excludedBin_;
        if (ebin > nbins_)
            ebin = nbins_;
        unsigned rebin = r.excludedBin_;
        if (rebin > r.nbins_)
            rebin = r.nbins_;

        return bandwidth_ == r.bandwidth_ &&
            degree_ == r.degree_ &&
            binwidth_ == r.binwidth_ &&
            nbins_ == r.nbins_ &&
            symbetaPower_ == r.symbetaPower_ &&
            bm_ == r.bm_ &&
            excludeCentralPoint_ == r.excludeCentralPoint_ &&
            ebin == rebin;
    }

    bool AllSymbetaParams1D::operator<(const AllSymbetaParams1D& r) const
    {
        if (bandwidth_ < r.bandwidth_) return true;
        if (bandwidth_ > r.bandwidth_) return false;
        if (degree_ < r.degree_) return true;
        if (degree_ > r.degree_) return false;
        if (binwidth_ < r.binwidth_) return true;
        if (binwidth_ > r.binwidth_) return false;
        if (nbins_ < r.nbins_) return true;
        if (nbins_ > r.nbins_) return false;
        if (symbetaPower_ < r.symbetaPower_) return true;
        if (symbetaPower_ > r.symbetaPower_) return false;
        if (bm_ < r.bm_) return true;
        if (r.bm_ < bm_) return false;
        if (!excludeCentralPoint_ && r.excludeCentralPoint_) return true;
        if (excludeCentralPoint_ && !r.excludeCentralPoint_) return false;

        unsigned ebin = excludedBin_;
        if (ebin > nbins_)
            ebin = nbins_;
        unsigned rebin = r.excludedBin_;
        if (rebin > r.nbins_)
            rebin = r.nbins_;
        return ebin < rebin;
    }
}
